import os
import json
from .omopcdm import OmopCDM

class MappingRules:

    def __init__(self, rulesfilepath):
        self.rules_data = self.load_json(rulesfilepath)
        self.omopcdm = OmopCDM()

    def load_json(self, f_in):
        """
        """
        if os.path.exists(f_in):
            data = json.load(open(f_in))
        else:
            try:
                data = json.loads(f_in)
            except Exception as err:
                raise FileNotFoundError(f"{f_in} not found. Or cannot parse as json")

        return data

    def get_all_outfile_names(self):
        file_list = []
        for outfilename in self.rules_data["cdm"]:
            file_list.append(outfilename)

        return file_list

    def parse_rules_src_to_tgt(self, infilename):
        """
        Parse rules to produce a map of source to target data for a given input file
        """
        outfilenames = []
        outdata = {}

        for outfilename, rules_set in self.rules_data["cdm"].items():
            for datatype, rules in rules_set.items():
                key, data = self.process_rules(infilename, outfilename, rules)
                if key != "":
                    if key not in outdata:
                        outdata[key] = []
                    outdata[key].append(data)
            if outfilename not in outfilenames:
                outfilenames.append(outfilename)

        return outfilenames, outdata

    def process_rules(self, infilename, outfilename, rules):
        outkey = ""
        data = {}
        plain_key = ""
        term_value_key = ""

        for outfield, source_info in rules.items():
            if source_info["source_field"] not in data:
                data[source_info["source_field"]] = []
            if source_info["source_table"] == infilename:
                if "term_mapping" in source_info:
                    if type(source_info["term_mapping"]) is dict:
                        for inputvalue, term in source_info["term_mapping"].items():
                            term_value_key = infilename + "~" + source_info["source_field"] + "~" + str(inputvalue) + "~" + outfilename
                            data[source_info["source_field"]].append(outfield + "~" + str(source_info["term_mapping"][str(inputvalue)]))
                    else:
                        plain_key = infilename + "~" + source_info["source_field"] + "~" + outfilename
                        data[source_info["source_field"]].append(outfield + "~" + str(source_info["term_mapping"]))
                else:
                    data[source_info["source_field"]].append(outfield)
        if term_value_key != "":
            return term_value_key, data

        return plain_key, data



class Metrics():
  def __init__(self):
    self.datasummary={}

  def add_data(self, desttablename, increment):
    """
    add_data(self, destination table, data increment)
    Apply the contents of a data increment to the stored self.datasummary
    """
    name = increment["name"]
    for datakey, dataitem in increment.items():
      if datakey == "valid_person_id":
        dkey = "NA" + "." + desttablename + "." + name + "." + datakey
        self.add_counts_to_summary(dkey, dataitem)
      elif datakey == "person_id":
        dkey = "NA" + "." + desttablename + "." + name + "." + datakey
        self.add_counts_to_summary(dkey, dataitem)
      elif datakey == "required_fields":
        for fieldname in dataitem:
          prfx = "NA"
          if "source_files" in increment:
            if fieldname in increment["source_files"]:
              prfx = self.get_prefix(increment["source_files"][fieldname]["table"])
          dkey = prfx + "." + desttablename + "." + name + "." + fieldname
          self.add_counts_to_summary(dkey, dataitem[fieldname])

  def get_prefix(self, fname):
    return fname.split(".")[0]

  def add_counts_to_summary(self, dkey, count_block):
    if dkey not in self.datasummary:
      self.datasummary[dkey] = {}
    for counttype in count_block:
      if counttype not in self.datasummary[dkey]:
        self.datasummary[dkey][counttype] = 0
      self.datasummary[dkey][counttype] += int(count_block[counttype])

  def get_summary(self):
    summary_str = "source\ttablename\tname\tcolumn name\tbefore\tafter content check\tpct reject content check\tafter date format check\tpct reject date format\n"
    for dkey in self.datasummary:
      #print(dkey)
      source, tablename, name, colname = dkey.split('.')
      before_count = int(self.datasummary[dkey]["before"])
      after_count = int(self.datasummary[dkey]["after"])
      after_pct = (float)(before_count - after_count) * 100 / before_count
      summary_str += source + "\t" + tablename + "\t" + name + "\t" + colname + "\t" + str(before_count) + "\t" + str(after_count) + "\t" + "{0:.3f}".format(after_pct) + "\t"
      if "after_formatting" in self.datasummary[dkey]:
        after_format_count = int(self.datasummary[dkey]["after_formatting"])
        after_format_pct = (float)(after_count - after_format_count) * 100 / after_count
        summary_str += str(after_format_count) + "\t" + "{0:.3f}".format(after_format_pct) + "\n"
      else:
        summary_str += "NA\tNA\n"
      #summary_str += "KEY {0}, COUNTS {1}\n".format(dkey, self.datasummary[dkey])

    return summary_str



class InputColumnData:

    def __init__(self):
        self.date_columns = {
            "gs_demog.csv": "yob",
            "gs_dolorisk.csv": "dolorisk_date",
            "gs_diabetes.csv": "diag_date",
            "gs_cases_icd10.csv": "case_date",
            "gs_cases_icd10_prev.csv": "appt",
            "gs_cases_icd9.csv": "case_date",
            "gs_cases_icd9_prev.csv": "appt",
            "gs_appt_events.csv": "appt",
            "SMR01_Demographics.csv": "dob",
            "SMR01_Condition.csv": "ADMISSION_DATE",
            "SMR01_Operations.csv": "DATE_OPERATION"
        }
        self.id_columns = {
            "gs_demog.csv": "id",
            "gs_dolorisk.csv": "id",
            "gs_diabetes.csv": "id",
            "gs_cases_icd10.csv": "id",
            "gs_cases_icd10_prev.csv": "id",
            "gs_cases_icd9.csv": "id",
            "gs_cases_icd9_prev.csv": "id",
            "gs_appt_events.csv": "id",
            "SMR01_Demographics.csv": "ENCRYPTED_UPI",
            "SMR01_Condition.csv": "ENCRYPTED_UPI",
            "SMR01_Operations.csv": "ENCRYPTED_UPI"
        }

    def is_data_column(self, colname, filename):
        if colname == self.date_columns[filename]:
            return False
        if colname == self.id_columns[filename]:
            return False
        return True

    def get_id_column(self, filename):
        if filename in self.id_columns:
            return self.id_columns[filename]
        return None

    def get_date_column(self, filename):
        if filename in self.date_columns:
            return self.date_columns[filename]
        return None



class OmopCDM:

    def __init__(self):
        self.all_columns = {
                         "condition_occurrence": ["condition_occurrence_id", "person_id",
                                                  "condition_concept_id", "condition_start_date",
                                                  "condition_start_datetime",    "condition_end_date",
                                                  "condition_end_datetime", "condition_type_concept_id",
                                                  "stop_reason", "provider_id", "visit_occurrence_id",
                                                  "condition_source_value", "condition_source_concept_id",
                                                  "condition_status_source_value", "condition_status_concept_id"],
                         "death": ["person_id", "death_date", "death_datetime", "death_type_concept_id",
                                   "cause_concept_id", "cause_source_value", "cause_source_concept_id"],
                         "drug_exposure": ["drug_exposure_id", "person_id", "drug_concept_id",
                                           "drug_exposure_start_date", "drug_exposure_start_datetime",
                                           "drug_exposure_end_date", "drug_exposure_end_datetime", "verbatim_end_date",
                                           "drug_type_concept_id", "stop_reason",
                                           "refills", "quantity", "days_supply", "sig", "route_concept_id",
                                           "lot_number", "provider_id", "visit_occurrence_id",
                                           "drug_source_value", "drug_source_concept_id", "route_source_value",
                                           "dose_unit_source_value"],
                         "measurement": ["measurement_id", "person_id", "measurement_concept_id", "measurement_date", "measurement_datetime", "measurement_type_concept_id",
                                         "operator_concept_id", "value_as_number", "value_as_concept_id", "unit_concept_id", "range_low", "range_high", "provider_id",
                                         "visit_occurrence_id", "measurement_source_value", "measurement_source_concept_id", "unit_source_value", "value_source_value"],
                         "observation": ["observation_id", "person_id", "observation_concept_id", "observation_date",
                                         "observation_datetime", "observation_type_concept_id", "value_as_number",
                                         "value_as_string", "value_as_concept_id", "qualifier_concept_id", "unit_concept_id provider_id",
                                         "visit_occurrence_id", "observation_source_value", "observation_source_concept_id", "unit_source_value", "qualifier_source_value"],
                         "person": ["person_id", "gender_concept_id", "year_of_birth", "month_of_birth", "day_of_birth", "birth_datetime", "race_concept_id", "ethnicity_concept_id",
                                    "location_id", "provider_id", "care_site_id", "person_source_value", "gender_source_value", "gender_source_concept_id",
                                    "race_source_value", "race_source_concept_id", "ethnicity_source_value", "ethnicity_source_concept_id"],
                         "procedure_occurrence": ["procedure_occurrence_id", "person_id",            "procedure_concept_id", "procedure_date",     "procedure_datetime", "procedure_type_concept_id", "modifier_concept_id",
                                                  "quantity", "provider_id", "visit_occurrence_id", "procedure_source_value", "procedure_source_concept_id", "qualifier_source_value"],
                         "specimen": ["specimen_id", "person_id", "specimen_concept_id", "specimen_type_concept_id", "specimen_date", "specimen_datetime", "quantity", "unit_concept_id",
                                      "anatomic_site_concept_id", "disease_status_concept_id", "specimen_source_id", "specimen_source_value", "unit_source_value", "anatomic_site_source_value", "disease_status_source_value"],
                         "visit_occurrence": ["visit_occurrence_id", "person_id", "visit_concept_id", "visit_start_date", "visit_start_datetime",
                                              "visit_end_date", "visit_end_datetime", "visit_type_concept_id", "provider_id", "care_site_id",
                                              "visit_source_value", "visit_source_concept_id", "admitting_source_concept_id", "admitting_source_value",
                                              "discharge_to_concept_id", "discharge_to_source_value", "preceding_visit_occurrence_id"]
                        }
        self.date_column_data = {
                "condition_occurrence": {"condition_start_datetime": "condition_start_date", "condition_end_datetime": "condition_end_date"},
                "death": {"death_datetime": "death_date"},
                "drug_exposure": {"drug_exposure_start_datetime": "drug_exposure_start_date", "drug_exposure_end_datetime": "drug_exposure_end_date"},
                "measurement": {"measurement_datetime": "measurement_date"},
                "observation": {"observation_datetime": "observation_date"},
                #"person": {"birth_datetime": ["year_of_birth", "month_of_birth", "day_of_birth"]},
                "procedure_occurrence": {"procedure_datetime": "procedure_date"},
                "specimen": {"specimen_datetime": "specimen_date"},
                "visit_occurrence": {"visit_start_datetime": "visit_start_date", "visit_end_datetime": "visit_end_date"}
                }
        self.datetime_columns = {
                "condition_occurrence": ["condition_start_datetime", "condition_end_datetime"],
                "death": ["death_datetime"],
                "drug_exposure": ["drug_exposure_start_datetime", "drug_exposure_end_datetime"],
                "measurement": ["measurement_datetime"],
                "observation": ["observation_datetime"],
                "person": ["birth_datetime"],
                "procedure_occurrence": ["procedure_datetime"],
                "specimen": ["specimen_datetime"],
                "visit_occurrence": ["visit_start_datetime", "visit_end_datetime"]
                }
        self.person_id_column = {
                "condition_occurrence": "person_id",
                "death": "person_id",
                "drug_exposure": "person_id",
                "measurement": "person_id",
                "observation": "person_id",
                "person": "person_id",
                "procedure_occurrence": "person_id",
                "specimen": "person_id",
                "visit_occurrence": "person_id"
                }
        self.auto_number_column = {
                "condition_occurrence": "condition_occurrence_id",
                "death": "death_id",
                "drug_exposure": "drug_exposure_id",
                "measurement": "measurement_id",
                "observation": "observation_id",
                "procedure_occurrence": "procedure_occurrence_id",
                "specimen": "specimen_id",
                "visit_occurrence": "visit_occurrence_id"
                }

    def get_column_map(self, colarr, delim=","):
        colmap = {}
        for i, col in enumerate(colarr):
            colmap[col] = i
        return colmap

    def get_omop_column_map(self, tablename):
        if tablename in self.all_columns:
            return self.get_column_map(self.all_columns[tablename])
        return None

    def get_omop_column_list(self, tablename):
        if tablename in self.all_columns:
            return self.all_columns[tablename]
        return None

    def get_omop_date_column_data(self, tablename):
        if tablename in self.date_column_data:
            return self.date_column_data[tablename]
        return {}

    def get_omop_datetime_fields(self, tablename):
        if tablename in self.datetime_columns:
            return self.datetime_columns[tablename]
        return None

    def get_omop_person_id_field(self, tablename):
        if tablename in self.person_id_column:
            return self.person_id_column[tablename]
        return None

    def get_omop_auto_number_field(self, tablename):
        if tablename in self.auto_number_column:
            return self.auto_number_column[tablename]
        return None
